from coproduct_recipes import Coproduct_Recipes
from recipe import Recipe
from item import Item

import jsonpickle as jp
from os import path

def save_recipes(recipes_list, filename='recipes.sc'):
    """
    Write Recipes to given filename for local storage.
    """
    # When file already present, wipe file, write new recipes
    if path.isfile(filename):
        print('Saved recipes file already exists. Overwriting.')
        erase_contents(filename)
        recipe_writer(filename, recipes_list, mode='a')

    # When file isn't present, make file, write new recipes
    else:
        print('Creating new recipes file for saving recipes.')
        recipe_writer(filename, recipes_list, mode='x')

def jsonify_recipes(coproduct_recipe_list):
    """
    Return Coproduct Recipe list encoded into a JSON format string.
    """
    return [jp.encode(coproduct_recipe) for coproduct_recipe in coproduct_recipe_list]

def recipe_writer(filename, recipes_list, mode):
    """
    Internal method for writing Recipes to given file.
    """
    try:
        # Write recipes to file in given mode
        with open(filename, mode) as recipes_file:
            for recipe_json in recipes_list:
                recipes_file.write(recipe_json)
                recipes_file.write('\n')

    except Exception:
        print('Problem writing to file.')
        # raise Exception

def erase_contents(filename):
    """
    Internal method for clearing the contents of given file.
    """
    try:
        # Clear contents of file
        open(filename, 'w').close()

    except Exception:
        print('Problem erasing file.')

def load_saved_recipes(filename):
    """
    Internal method for reading contens of given file.
    """
    contents = []
    
    try:
        # Read file line by line, recipe by recipe
        with open(filename, 'r') as recipes_file:
            contents = recipes_file.readlines()
        
        return contents

    except Exception:
        print('Problem reading recipes from file.')

def convert_loaded_recipes(loaded_recipes_list):
    """
    Return encoded recipes as list of Coproduct Recipe instances.
    """
    return [jp.decode(recipe) for recipe in loaded_recipes_list]

def save_recipes_locally(recipes_list, filename='recipes.sc'):
    """
    Write scraped recipes to given filename.
    """
    # Convert Coproduct_Recipes to JSON serializable
    jsonified_recipes = jsonify_recipes(recipes_list)

    # Write Corpoduct_Recipes to file
    save_recipes(jsonified_recipes, filename)

def read_saved_recipes(filename='recipes.sc'):
    """
    Read given filename to get all saved Coproduct Recipes instances.
    """
    # Read saved recipes from filename
    loaded_recipes = load_saved_recipes(filename)
    
    # Deserialize recipes to read in as Coproduct Recipes class
    loaded_recipes_deserialized = convert_loaded_recipes(loaded_recipes)

    return loaded_recipes_deserialized



from item import Item
from recipe import Recipe
from coproduct_recipes import Coproduct_Recipes

all_requirements = []

class CraftingTree():

    def __init__(self, product_name: str, product_rate: float, coproduct_recipes_list: list):
        """
        Initialize tree with target item name and desired production rate.
        """
        init_item = Item(product_name, 'solid', product_rate)
        self._coproduct_recipes_list = coproduct_recipes_list
        self._root_node = CraftingNode(init_item, self._coproduct_recipes_list, 0)
        # self._tree_height = 0
        # self._all_reqs = []

        self.calc()

    @property
    def root(self):
        return self._root_node

    @root.setter
    def root(self, root):
        self._root_node = root

    @property
    def recipes_list(self):
        return self._coproduct_recipes_list

    @root.setter
    def recipes_list(self, recipes_list):
        self._coproduct_recipes_list = recipes_list
    
    @property
    def tree_height(self):
        return self._tree_height
    
    @property
    def reqs(self):
        return self._all_reqs

    def calc(self):
        self.root.calc()

    def traverse(self):
        self.root.traverse()


class CraftingNode():

    MAX_CHILDREN = 4

    def __init__(self, init_item: Item, coproduct_recipes_list: list, level: int):
        """
        Initialize node using Item class definition.
        """
        self.coproduct_recipes_list = coproduct_recipes_list
        self.num_buildings = 0.0
        self.goal = init_item
        self.byproducts = []
        self.children = []
        self.level = level
        self.recipe = ''

    def __str__(self):
        return('Crafting node for {} at {}/min.'.format(self.goal.name, self.goal.rate))

    def __repr__(self):
        return('CraftingNode(Item({}, \'solid\', {}))'.format(self.goal.name, self.goal.rate))

    def display(self):
        """
        Print summary of CraftingNode, as part of CraftingTree.
        """
        # Set spacer strings for the following text
        align_str_header = '\t'*self.level
        align_str_body = '\t'*self.level
        
        # Exception for first node printed (root)
        if self.level > 0:
            align_str_header = '\t'*self.level + 'L_'
            align_str_body = '\t'*(self.level) + '  '

        # Display input names and required rates
        self.display_header(align_str_header)

        # Display names and rates of byproducts if any
        self.display_byproduct(align_str_body)

        # Display building / to extract
        self.display_building(align_str_body)

        # Print whitespace separator
        self.spacer()

    def display_header(self, align_str):
        """
        Print header of current CraftingNode.
        """
        print(align_str, self.unpack_list())

    def display_byproduct(self, align_str):
        """
        Print byproduct section of current CraftingNode.
        """
        if self.byproducts:
            print(align_str, 'Byproduct: {}'.format(self.byproducts[0]))

    def display_building(self, align_str):
        """
        Print building section of current CraftingNode.
        """
        if self.recipe:
            if self.recipe._building:
                print(align_str, 'Made in {:.2f}x {}.'.format(self.num_buildings, self.recipe.building))
            else:
                raise Exception('Recipe exists with no building.')
        else:
            print(align_str, 'Mine/Extract resource from the world.')

    def spacer(self):
        print()

    def unpack_list(self):
        """
        Return appropriate print method from Item instace of self.goal.
        """
        return self.goal.__str__()

    def user_recipe_select(self, potential_recipes):
        """
        Offer user choice of which recipe to use for the crafting step.
        """
        # Check for edge cases
        if potential_recipes.num_recipes == 0:
            return None

        if potential_recipes.num_recipes == 1:
            return potential_recipes.recipes[0]
        
        # If no edge case, display options
        potential_recipes.print_summary()

        # Prompt user for resonable selection
        try:
            selected_idx = int(input('Enter recipe # to use: '))
        except ValueError as err:
            selected_idx = int(input('Enter valid recipe # to use: '))
        
        self.spacer()
        while selected_idx > potential_recipes.num_recipes or selected_idx < 0:
            selected_idx = int(input('Enter valid recipe # to use: '))

        # Return the Recipe at the adjusted index (1-indexed)
        return potential_recipes.recipes[selected_idx-1]
    
    @property
    def num_children(self):
        """
        Return number of children a node already has.
        """
        return len(self.children)

    def add_child(self, new_child):
        """
        Add a child to node. Handles max amount of children already existing.
        """
        if self.num_children < self.MAX_CHILDREN:
            self.children.append(new_child)
        
        else:
            raise Exception('Max number of allowed children ({}) exceeded.'.format(self.MAX_CHILDREN))
    
    def get_suitable_recipes(self, desired_output):
        """
        Return the CRs instance containing Recipes for the desired output.
        """
        for coproduct_recipe in self.coproduct_recipes_list:
            if desired_output == coproduct_recipe.product:
                return coproduct_recipe
        
        return None
        
    def calc(self):
        """
        Take root node of tree, with user assistance, compute the crafting requirements and build the tree.
        """
        # Find appropriate recipes
        suitable_recipe = self.get_suitable_recipes(self.goal.name)

        # If recipe for desired goal exists
        if suitable_recipe:

            # If a non-raw recipe has been returned, calculate production requirements
            if suitable_recipe.recipes:

                # Ask user for which recipe to use
                self.recipe = self.user_recipe_select(suitable_recipe)
                all_inputs = self.recipe.input_items
                all_output_names = self.recipe.output_names

                # Build crafting tree for each input for recipe
                for inp in all_inputs:

                    # Calculate the requirements for the step
                    multiplier = self.recipe.get_ratio(inp.name, self.goal.name)

                    # Calculate number of buildings required for the step
                    self.num_buildings = self.goal.rate / self.recipe.get_output_rate_for(self.goal.name)

                    # Set goal for next child
                    next_goal = Item(inp.name, inp.form, self.goal.rate * multiplier)

                    # Add next steps as child CraftingNode
                    self.add_child(CraftingNode(next_goal, self.coproduct_recipes_list, self.level+1))

                # Find and report byproducts for the step if present
                if len(all_output_names) > 1:

                    # Remove step goal from outputs list
                    all_output_names.remove(self.goal.name)

                    # Convert output names to output Item instances
                    all_output_items = [self.recipe.get_output_item_from_name(output) for output in all_output_names]

                    # Scale byproduct rates appropriately
                    self.byproducts = [output.scale_by(self.num_buildings) for output in all_output_items]

                # Recurse
                for child in self.children:
                    child.calc()

            # Otherwise, end branch
            else:
                # print('Reached leaf node.')
                all_requirements.append(self.goal)

        # When no recipe for requested component exists
        else:
            raise Exception('No match for desired output in any available recipe.', self.goal.name)

    def traverse(self):
        """
        Print CraftingNodes as tree is being traversed depth-first.
        """
        self.display()

        for child in self.children:
            child.traverse()




class Item():
    """Any Extracted item or item made in a production building"""

    def __init__(self, name: str, form: str, rate: float):
        self._name = name
        self._form = form
        self._rate = rate

    def __str__(self):
        return f'{self._rate}/min {self._name}.'

    def __repr__(self):
        return f'Item(name=\'{self._name}\', form=\'{self._form}\', rate={self._rate})'
        
    # For when/if form is added
    #return 'Item(name=\'{}\', form=\'{}\', rate={})'.format(self.name, self.form, self.rate)

    def __add__(self, other):
        """
        Return the combined Item instance.
        """
        return Item(self._name, self._form, self._rate + other._rate)
    
    @property
    def name(self):
        return self._name

    @name.setter
    def name(self, name: str):
        self._name = name

    @property
    def form(self):
        return self._form

    @form.setter
    def form(self, form: str):
        self._form = form

    @property
    def rate(self):
        return float(self._rate)

    @rate.setter
    def rate(self, rate: float):
        self._rate = rate

    @property
    def is_solid(self):
        return self._form == 'solid'

    @property
    def is_liquid(self):
        return self._form == 'liquid'
    
    @property
    def is_gas(self):
        return self._form == 'gas'

    def scale_by(self, multiplier: float):
        return Item(self._name, self._form, self._rate * multiplier)

    # def simple_str(self):
    #     return '{}x {}'.format(self._rate, self._name)

# test_item = Item('test_item', 'solid', 120)
# print(test_item)



class Building():
    """Building with associated number of inputs and outputs, power consumption, overclock level"""
    name = ''
    input_num = 0
    output_num = 0
    base_power_use = 0
    overclock = 0

    def __init__(self, name='', input_num=0, output_num=0, base_power_use=0, overclock=0):
        """Read in from file, set building stats"""
        self.name = name
        self.input_num = input_num
        self.output_num = output_num
        self.base_power_use = base_power_use
        self.overclock = overclock

    def __str__(self):
        return '{} with {} inputs, {} outputs, {} MW of power use and overclock level of {}'.\
            format(self.name, self.input_num, self.output_num, self.base_power_use, self.overclock)
    
    def __repr__(self):
        return 'Building(name=\'{}\', input_num={}, output_num={}, base_power_use={}, overclock={})'.\
            format(self.name, self.input_num, self.output_num, self.base_power_use, self.overclock)

    def is_extactor(self):
        return self.input_num == 0

    def is_producer(self):
        return self.input_num != 0
    
    def get_base_power_use(self):
        return self.base_power_use

    def get_net_power_use(self):
        return self.base_power_use * self.get_power_multiplier()

    def get_overclock(self):
        return self.overclock

    def get_power_multiplier(self):
        # Needs actual formula, overclock power scaling is not linear
        return 1 + self.overclock*.5


from recipe import Recipe
from item import Item

class Coproduct_Recipes():
    """
    Combines Recipe instances and common product name string.
    """

    def __init__(self, product_name: str, recipes_list: list):
        """
        Initialize instance with product name string and list of Recipe instances.
        """
        self._product_name = product_name
        self._recipes_list = recipes_list

    def __str__(self):
        """
        Return summary of instance using product name and number of recipes included.
        """
        num_recipes = self.num_recipes
        string = '{}: {} recipe'.format(self._product_name, num_recipes)

        if num_recipes > 1:
            return string + 's'
        else:
            return string
    
    def __repr__(self):
        """
        TODO
        """
        return self.__str__()

    def print_summary(self):
        """
        Prints all Recipe instances contained within Coproduct Recipe instance.
        """
        n = 1
        for recipe in self.recipes:
            print(str(n), '-', recipe.summary)
            n += 1

    @property
    def product(self):
        """
        Return product name string.
        """
        return self._product_name

    @product.setter
    def product(self, product):
        """
        Product name property setter.
        """
        self._product_name = product

    @property
    def recipes(self):
        """
        Return list of all included Recipe instances.
        """
        return self._recipes_list

    @recipes.setter
    def recipes(self, recipes):
        """
        Recipes property setter.
        """
        self._recipes = recipes

    @property
    def num_recipes(self):
        """
        Return number of included Recipe instances.
        """
        return len(self._recipes_list)

    @property
    def is_raw(self):
        """
        Return True if material is raw.
        """
        return self._recipes_list == []

    # def get_base_recipes(self):
    #     """
    #     Return list of all included base Recipe instances.
    #     TODO fix
    #     """
    #     return [recipe for recipe in self.recipes_list if not recipe.is_alternate()]
    
    # def get_alt_recipes(self):
    #     """
    #     Return list of all included alternate Recipe instances.
    #     TODO fix
    #     """
    #     return [recipe for recipe in self.recipes_list if recipe.is_alternate()]

    # def __len__(self):
    #     """
    #     TODO
    #     """
    #     return(self.num_recipes)

    # def __getitem__(self, key):
    #     return(self.recipes_list[key])


#!/usr/bin/python

from recipe_storage import read_saved_recipes, load_saved_recipes, save_recipes_locally
from fetch_recipes import get_all_coproduct_recipes
from fetch_item_names import get_all_item_URLs
from tree import CraftingTree

import sys

def main():

    def help():
        HELP_STRING = ('Satisfy-calc use guide:\n'
                    '1. "update" or -u: to update local recipe files. Advisable after an update drops.\n'
                    '2. "calc" or -c: followed by \'Item Name\' Item Rate to calculate crafting chain for an ingredient. '
                    'e.g. ./satisfy-calc.py -c \'Iron Ingot\' 60\n'
                    '3. "help" or -h: to bring up this information.')
        print(HELP_STRING)

    def update_local():
        all_item_URLs = get_all_item_URLs()
        all_recipes = get_all_coproduct_recipes(all_item_URLs)
        save_recipes_locally(all_recipes)

    def calc():
        CRs = read_saved_recipes('recipes.sc')
        tree = CraftingTree(*args, CRs)
        tree.traverse()

    opts = [opt for opt in sys.argv[1:] if opt.startswith('-')]
    args = [arg for arg in sys.argv[1:] if not arg.startswith('-')]

    if 'help' in args or '-h' in opts:
        help()
    elif 'update' in args or '-u' in opts:
        print('Updating locally saved recipes... ')
        update_local()
    elif 'calc' in args or '-c' in opts:
        print('Begin crafting tree calculation... \nPlease select what recipe to use at each step: \n')
        calc()
    else:
        raise SystemExit('Try -h or --help.')

if __name__ == '__main__':
    main()
    



from bs4 import BeautifulSoup
from building import Building
import requests

# URL = 'https://satisfactory.gamepedia.com/Manufacturer'
# page = requests.get(URL)

# soup = BeautifulSoup(page.content, 'html.parser')

# print(soup.prettify())

# inputs = soup.find('b', string='Inputs')
# outputs = soup.find('b', string='Outputs')
# power = soup.find('b', string='Power Usage')

# values = soup.find_all(class_='infobox-row-value')
# for result in results:
#     print(result.prettify())

# for label in labels:
#     print(label)
# print('###############')
# for value in values:
#     print(value)

# for label in labels[-7:-3]:
#     if label.text.strip() == 'Outputs':
#         print(label.text.strip())

# for value in values[-7:-3]:
#     print((value.text).strip())

# print(inputs.parent.parent.find('td').text.strip())
# print(outputs.parent.parent.find('td').text.strip())
# print(power.parent.parent.find('td').text.strip())

# Need to dynamically form this list from a prior scrape
buildings = ['Manufacturer', 'Assembler', 'Constructor']

# Import from a config file maybe?
base_URL = 'https://satisfactory.fandom.com/'

for building in buildings:
    building_page = requests.get(base_URL + building)
    building_soup = BeautifulSoup(building_page.content, 'html.parser')

    # Good enough for basic stuff, needs to be able to handle lists for refineries and packagers
    inputs = building_soup.find('b', string='Inputs')
    outputs = building_soup.find('b', string='Outputs')
    power = building_soup.find('b', string='Power Usage')

    if None in (inputs, outputs, power):
        continue
    
    # print(inputs.parent.parent.find('td').text.strip())
    # print(outputs.parent.parent.find('td').text.strip())
    # print(power.parent.parent.find('td').text.strip())

    test = Building(name=building, 
                input_num= inputs.parent.parent.find('td').text.strip(),
                output_num= inputs.parent.parent.find('td').text.strip(),
                base_power_use= inputs.parent.parent.find('td').text.strip(),
                overclock=0)

    print(test)





from coproduct_recipes import Coproduct_Recipes as CR
from building import Building
from recipe import Recipe
from item import Item

from bs4 import BeautifulSoup
from bs4 import SoupStrainer
import requests
import re

def get_recipe_name(table_td):
    """
    Return name of recipe from Soup.
    """
    return str(table_td.contents[0])

def get_recipe_rate(string):
    """
    Return item input/output rate from string.
    """
    return float(re.match('[0-9]*[.]*[0-9]*', string).group(0))

def get_input_output(table_td):
    """
    Return list of items and rates from given table.
    """
    items_list = []
    rates_list = []

    while table_td:
        try:
            items_list.append(table_td.div.div.a.get('title'))
            rates_list.append(get_recipe_rate(table_td.div.next_sibling.string))
            
        except:
            break
        table_td = table_td.next_sibling

    return [items_list, rates_list]

def get_recipe_inputs(table_td):
    """
    Return list of all input names, list of all input rates.
    """
    return get_input_output(table_td)

def get_recipe_outputs(table_td):
    """
    Return list of all output names, list of all output rates.
    """
    return get_input_output(table_td)

def get_building(table_td):
    """
    Return the building recipe is made in.
    """
    try:
        return table_td.span.a.get('title')
    
    except:
        return None

def get_prereq(table_td):
    """
    Return what prerequisite exists for recipe unlock.
    """
    return str(table_td.span.text)

def get_is_alternate(table_td):
    """
    Return if the recipe is an alternate recipe (locked by default).
    """
    return True if table_td.br else False

def extract_product(extension: str):
    """
    Return full product name from URL extension.
    """
    return re.search('[a-zA-Z_-]+$', extension)[0]

def get_recipe_rows(recipe_soup):
    """
    Return non-header rows from crafting table on the item's Wiki page soup.
    """
    # Get all rows for recipe page
    try:
        crafting_table = recipe_soup.find_all(class_='wikitable')[0]
    except:
        raise Exception('Problem parsing wiki page soup.')

    crafting_rows = crafting_table.find_all('tr')

    # Select all non-header rows
    non_header_rows = crafting_rows[1:]

    return non_header_rows

def get_all_URLs_to_scrape(extensions_list):
    """
    Return list of full item URLs for scraping.
    """   
    base_URL = 'https://satisfactory.fandom.com'

    return [base_URL + extension for extension in extensions_list]

def itemify_components(names, rates):
    """
    Return Item instances list given names and rates.
    TODO: include other item forms besides 'solid' if necessary
    """
    return [Item(i, 'solid', r) for i, r in zip(names, rates)]

def scrape_recipe_page(recipes_soup):
    """
    Return Coproduct Recipes list for given Wiki page Soup.
    """
    # Select all non-header rows
    select_rows = get_recipe_rows(recipes_soup)

    # Define function-internal lists
    final_recipe_list = []
    temp_inputs = []
    temp_irates = []
    temp_outputs = []
    temp_orates = []

    # For each row in table (besides column names)
    for i in range(len(select_rows)):

        # Get first row's input item names and rates listed on HTML structure with 'class' attr
        if select_rows[i].has_attr('class'):

            # Get name for recipe, always index 0
            recipe_name = get_recipe_name(select_rows[i].find_all('td')[0])

            # Get if recipe is alt, always index 0
            recipe_is_alt = get_is_alternate(select_rows[i].find_all('td')[0])

            # Get primary row input values. always index 1+
            inps, irts = get_recipe_inputs(select_rows[i].find_all('td')[1])
            temp_inputs = inps
            temp_irates = irts

            # Get proper index for building info location in row
            if len(temp_inputs) < 2:
                building_index = 2

            else:
                building_index = 3

            # If exists, get first row's secondary inputs listed on HTML structure with no 'class' attr
            try:
                if not select_rows[i+1].has_attr('class'):

                    # Append secondary row input values to primary ones
                    inps, rts = get_recipe_inputs(select_rows[i+1].find_all('td')[0])
                    temp_inputs += inps
                    temp_irates += rts

            except:
                pass

            # Get building info
            recipe_building = get_building(select_rows[i].find_all('td')[building_index])

            # Get outputs for recipe, one index up from building
            outs, orts = get_recipe_outputs(select_rows[i].find_all('td')[building_index+1])
            temp_outputs = outs
            temp_orates = orts

            # Get prereqs for recipe
            # recipe_prereqs = get_prereq()
        
        # If hit row with no primary values, skip
        else:
            continue

        # Arrange items and rates into correct classes
        inputs = itemify_components(temp_inputs, temp_irates)
        outputs = itemify_components(temp_outputs, temp_orates)

        # Add to list of recipes for the page
        final_recipe_list.append(Recipe(recipe_name, inputs, outputs, recipe_building, recipe_is_alt))

    return final_recipe_list

def underscore_2_space(string: str):
    """
    Return string with underscores replaced by spaces.
    """
    return re.sub('[_]', ' ', string)

def get_recipe_soup(recipe_URL: str):
    """
    Return Soup of given URL.
    """
    # Get HTML of recipe page
    recipe_page = requests.get(recipe_URL)
    
    # Only parse and soup-ify the recipes table
    only_recipe_table = SoupStrainer(class_='wikitable')

    recipe_soup = BeautifulSoup(recipe_page.content, 'html.parser', parse_only=only_recipe_table)

    return recipe_soup

def get_all_coproduct_recipes(extensions_list: list):
    """
    Return Coproduct Recipes list for all recipes on the Wiki.
    """
    # Get list of item name URLs, convert to full URLs for scraping.
    all_item_URLs = get_all_URLs_to_scrape(extensions_list)

    # Get Soup list from all item URLs.
    all_recipe_soups = [get_recipe_soup(recipe_page) for recipe_page in all_item_URLs]

    # Scrape recipes from all Soups.
    all_recipe_lists = [scrape_recipe_page(recipe_soup) for recipe_soup in all_recipe_soups]

    # Make CR list from extensions and Recipe instances.
    all_coproduct_recipes = [CR(underscore_2_space(extract_product(extension)), recipe_list) for extension, recipe_list in zip(extensions_list, all_recipe_lists)]

    # Add Water, Fuel as a resource because it is missing
    all_coproduct_recipes.append(CR('Water', []))

    return all_coproduct_recipes



import pathlib
from setuptools import setup

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README file
README = (HERE / "README.md").read_text()

setup(
        name="satisfy-calc",
        version="1.1.2",
        description="Command line crafting tree visualizer for Satisfactory Game by CoffeeStain Studios",
        long_description=README,
        long_description_content_type="text/markdown",
        url="https://github.com/sedatDemiriz/satisfy-calc",
        author="Sedat Demiriz",
        author_email="sedatdemiriz97@gmail.com",
        license="MIT",
        classifiers=[
            "License :: OSI Approved :: MIT License",
            "Programming Language :: Python :: 3",
            "Programming Language :: Python :: 3.7",
        ],
        install_requires=["bs4", "jsonpickle", "requests"],
        packages=["satisfy_calc"],
        entry_points={
            "console_scripts": [
                "satisfy_calc=satisfy_calc.__main__:main"
            ]
        },
    )

# setup(
#     name="realpython-reader",
#     version="1.0.0",
#     description="Read the latest Real Python tutorials",
#     long_description=README,
#     long_description_content_type="text/markdown",
#     url="https://github.com/realpython/reader",
#     author="Real Python",
#     author_email="info@realpython.com",
#     license="MIT",
#     classifiers=[
#         "License :: OSI Approved :: MIT License",
#         "Programming Language :: Python :: 3",
#         "Programming Language :: Python :: 3.7",
#     ],
#     packages=["reader"],
#     include_package_data=True,
#     install_requires=["feedparser", "html2text"],
#     entry_points={
#         "console_scripts": [
#             "realpython=reader.__main__:main",
#         ]
#     },
# )




from item import Item
from re import sub

class Recipe():
    """
    Combines up to 4 Item instances to output up to 2 Item instances.
    """

    def __init__(self, recipe_name: bool, inputs_list: list, outputs_list: list, building: str, is_alternate: bool):
        """
        Initialize instance with name string, input Item instances list, 
        output Item instances list, building name, whether is an alternate recipe.
        """
        self._recipe_name = recipe_name
        self._inputs = inputs_list
        self._outputs = outputs_list
        self._building = building
        self._is_alternate = is_alternate

    def __str__(self):
        """
        Return all details of Recipe.
        """
        return f'Recipe name: {self.name}\nInputs: {self.input_names}\nOutputs: {self.output_names}\nMade in: {self.building}\nIs alt. recipe: {self.is_alt}\n'
    def __repr__(self):
        """
        Return Recipe name.
        """
        return f'Recipe for {self.name}'

    def scale_by(self, multiplier):
        """
        Scale inputs and outputs according to multiplier.
        """
        self.scale_inputs_by(multiplier)
        self.scale_outputs_by(multiplier)

    def scale_inputs_by(self, multiplier):
        """
        Scale inputs of recipe according to multiplier.
        """
        self._inputs = [inp.scale_by(multiplier) for inp in self._inputs]

    def scale_outputs_by(self, multiplier):
        """
        Scale inputs of recipe according to multiplier.
        """
        self._outputs = [outp.scale_by(multiplier) for outp in self._outputs]

    @property
    def summary(self):
        """
        Return summary of the Recipe instance.
        """
        return '[{}] made from {}'.format(self.name, self.list_formatter(self.input_names))

    def list_formatter(self, l):
        return '[' + ', '.join(l) + ']'

    @property
    def name(self):
        """
        Return Recipe instance name.
        """
        return self._recipe_name

    @name.setter
    def name(self, name):
        self._recipe_name = name
    
    @property
    def is_raw(self):
        """
        Return True if the item is a raw material (not craftable/no inputs).
        """
        return self._inputs is None

    @property
    def input_items(self):
        """
        Return all inputs as list of Item instances.
        """
        return self._inputs

    @input_items.setter
    def input_items(self, input_items):
        self._inputs = input_items

    @property
    def input_names(self):
        """
        Return all inputs as list of item name strings.
        """
        return [inp.name for inp in self._inputs]

    @property
    def input_rates(self):
        """
        Return all inputs as list of item rates.
        """
        return [inp.rate for inp in self._inputs]

    @property
    def output_items(self):
        """
        Return all outputs as list of Item instances.
        """
        return self._outputs

    @output_items.setter
    def output_items(self, output_items):
        self._outputs = output_items

    @property
    def output_names(self):
        """
        Return all outputs as list of item names.
        """
        return [outp.name for outp in self._outputs]

    @property
    def output_rates(self):
        """
        Return all outputs as list of item rates.
        """
        return [outp.rate for outp in self._outputs]

    @property
    def building(self):
        """
        Return building name.
        """
        return self._building

    @building.setter
    def building(self, building):
        self._building = building

    @property
    def is_base(self):
        """
        Return whether recipe is unlocked by default.
        """
        return self._is_alternate == False

    @property
    def is_alt(self):
        """
        Return whether recipe is locked by default.
        """
        return self._is_alternate

    @is_alt.setter
    def is_alt(self, is_alternate):
        self._is_alternate = is_alternate

    def get_output_rate_for(self, output_name):
        return self.output_rates[self.output_names.index(output_name)]
        
    def get_input_item_from_name(self, item_name: str):
        """
        Return Item instance from self.inputs given item name.
        """
        return [inp for inp in self.input_items if item_name == inp.name][0]

    def get_output_item_from_name(self, item_name: str):
        """
        Return Item instance from self.outputs given item name.
        """
        return [outp for outp in self.output_items if item_name == outp.name][0]

    def get_ratio(self, inp: str, outp: str):
        """
        Return input/output ratio given one input, one output item name.
        """
        if inp is None:
            return None
        else:
            return self.get_input_item_from_name(inp).rate /\
                self.get_output_item_from_name(outp).rate
    
    def get_ratio_rev(self, inp: str, outp: str):
        """
        Return output/input ratio given one input, one output item name.
        """
        if inp is None:
            return None
        else:
            return self.get_output_item_from_name(outp).rate /\
                self.get_input_item_from_name(inp).rate



from bs4 import BeautifulSoup
from bs4 import SoupStrainer

import requests
import re

def get_item_names_from_url(URL, id_tag='mw-pages'):
    """
    Returns a Soup of the Items page from the Satisfactory Wiki.
    """
    # Navigate to source URL
    items_page = requests.get(URL)

    # Setup Strainer for specific tag
    only_item_names = SoupStrainer(id=id_tag)

    # Extract html within the id_tag as Soup
    items_soup = BeautifulSoup(items_page.content, 'html.parser', parse_only=only_item_names)
    
    return items_soup

def get_list_of_item_names(items_soup):
    """
    Returns list of URL extensions for all items in the game.
    """
    # Find all alphabetical listings for items
    all_item_tags = items_soup.find_all('li')

    # Extract link locations for all items
    all_item_names = [i.a.get('href') for i in all_item_tags]

    return all_item_names

def filter_item_names(items_URL_extensions, filter_string='/[a-zA-Z]{2}$|^/%'):
    """
    Returns scraped URL extensions after filtering out non-crafting related items.
    TODO: Have these items in a separate file in the future.
    """

    # List of URLs to explicitly exclude
    blacklist = [
                    '/wiki/Cup', 
                    '/wiki/Vines', 
                    '/wiki/Statue',
                    '/wiki/HUB_Parts', 
                    '/wiki/Hard_Drive',
                    '/wiki/Somersloop', 
                    '/wiki/FICSIT_Coupon', 
                    '/wiki/Mercer_Sphere',
                    '/wiki/Quantum_Computer',
                    '/wiki/Superposition_Oscillator',
                    '/wiki/%EC%B2%A0_%EC%A3%BC%EA%B4%B4',
                    '/wiki/User:SignpostMarv/Draft/Hard_Drive', 
                ]

    # List of URLs that are not on the all_items_URL (All craftable fluids atm.)
    whitelist = [
                    '/wiki/Fuel',
                    '/wiki/Turbofuel',
                    '/wiki/Nitric_Acid',
                    '/wiki/Nitrogen_Gas',
                    '/wiki/Sulfuric_Acid',
                    '/wiki/Liquid_Biofuel',
                    '/wiki/Alumina_Solution',
                    '/wiki/Heavy_Oil_Residue',
                ]

    filtered_items_URL_extensions = [item_name for item_name in items_URL_extensions if re.search(filter_string, item_name) is None]
    filtered_items_URL_extensions = [item_name for item_name in filtered_items_URL_extensions if item_name not in blacklist]
    filtered_items_URL_extensions = filtered_items_URL_extensions + whitelist

    return filtered_items_URL_extensions

def get_all_item_URLs():
    """
    Returns all available items' URL extensions available on the Satisfactory Wiki.
    """
    # Specify URL to fetch item names from
    all_items_URL = 'https://satisfactory.gamepedia.com/Category:Items'

    # Get soup of specified URL
    items_page_soup = get_item_names_from_url(all_items_URL, 'mw-pages')

    # Get list of item URL extensions from Soup
    items_URL_extensions = get_list_of_item_names(items_page_soup)

    # Filter item URLs list to exclude problematic instances
    filtered_items_URL_extensions = filter_item_names(items_URL_extensions)

    return filtered_items_URL_extensions
